# -*- coding: utf-8 -*-
"""Bottle.py middleware development utilities.

Provides a basic interface for applying a middleware to 'bottle.py' apps or web
services.
"""

from __future__ import absolute_import


__author__ = 'pav'
__date__ = '18-04-2015'
__all__ = ['BaseMiddleware', 'StripPathMiddleware']


from bottle_neck import __version__

version = tuple(map(int, __version__.split('.')))


class BaseMiddleware(object):
    """BaseMiddleware Abstract class. Provides an interface for creating
    middleware classes that act like proxy-wrappers to the actual bottle
    application instance.

    Subclass BaseMiddleware class and implement `__call__(self, e, h)`
    magic method in order to create an application transparent middleware


    Examples:
        >>> import bottle
        >>>
        ... class MyMiddle(BaseMiddleware):
        ...     def __call__(self, e, h):
        ...         print 'Hi from middleware'
        ...         self.app(e, h)
        ...
        >>> app = MyMiddle(bottle.Bottle())
        >>> print hasattr(app, 'route')
        True
        >>> class MyMw1(BaseMiddleware):
        ...     def __call__(self, e, h):
        ...         print 'Hi from middleware 1'
        ...         self.app(e, h)
        ...
        >>> class MyMw2(BaseMiddleware):
        ...     def __call__(self, e, h):
        ...         print 'Hi from middleware 2'
        ...         self.app(e, h)
        ...
        >>> app = MyMw2(MyMw1(bottle.Bottle()))
        >>> print hasattr(app, 'route')
        True
    """

    def __init__(self, app):
        self.app = app

    def __getattr__(self, attr):
        try:
            return getattr(self.app, attr)
        except AttributeError:  # pragma: no cover
            return None

    def __dir__(self):  # pragma: no cover
        return dir(self.app)

    def __call__(self, e, h):  # pragma: no cover
        raise NotImplementedError('Must implement __call__ method')


class StripPathMiddleware(BaseMiddleware):
    """Middleware class that handles url ending slashes.

    Development does not enforce URL dispatching with trailing slashes,
    but the end user has both kind of `variations` available.
    Straight copy/pasted from 'bottle.py' docs.
    """

    def __call__(self, e, h):  # pragma: no cover
        e['PATH_INFO'] = e['PATH_INFO'].rstrip('/')
        return self.app(e, h)


if __name__ == '__main__':  # pragma: no cover
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
"""Bottle.py routing development utilities.

Provides a basic routing mechanism for 'bottle.py' web services, with a simple
and unique interface for function-based / Class-based handlers.
"""

from __future__ import absolute_import

__author__ = 'pav'
__date__ = '18-04-2015'
__all__ = ['Router', 'RouteError']


from bottle_neck import __version__
from bottle_neck.cbv import HandlerMeta
import types

version = tuple(map(int, __version__.split('.')))


class RouteError(Exception):
    """Raise when a routing error occurs.
    """
    pass


class Route(object):
    """Base Route interface.

    It wraps function-based web handlers in order to provide a
    same interface for `bottle_neck.routing.Router` class functionality.
    """

    __slots__ = ('uri', 'methods', 'callable_obj')

    @classmethod
    def wrap_callable(cls, uri, methods, callable_obj):
        """Wraps function-based callable_obj into a `Route` instance, else
        proxies a `bottle_neck.handlers.BaseHandler` subclass instance.

        Args:
            uri (str):  The uri relative path.
            methods (tuple): A tuple of valid method strings.
            callable_obj (instance): The callable object.

        Returns:
            A route instance.

        Raises:
            RouteError for invalid callable object type.
        """
        if isinstance(callable_obj, HandlerMeta):
            callable_obj.base_endpoint = uri
            callable_obj.is_valid = True
            return callable_obj

        if isinstance(callable_obj, types.FunctionType):
            return cls(uri=uri, methods=methods, callable_obj=callable_obj)

        raise RouteError("Invalid handler type.")

    def __init__(self, uri, methods, callable_obj):
        self.uri = uri,
        self.methods = methods,
        self.callable_obj = callable_obj

    @property
    def is_valid(self):
        args = [self.uri, self.methods, self.callable_obj]
        return all([arg for arg in args])

    def register_app(self, app):
        """Register the route object to a `bottle.Bottle` app instance.

        Args:
            app (instance):

        Returns:
            Route instance (for chaining purposes)
        """
        app.route(self.uri, methods=self.methods)(self.callable_obj)

        return self


class Router(object):
    """Base Router class for bottle.py WSGI applications.
    """

    def __init__(self):
        self._routes = set()

    @property
    def routes(self):  # pragma: no cover
        return self._routes

    def register_handler(self, callable_obj, entrypoint, methods=('GET',)):
        """Register a handler callable to a specific route.

        Args:
            entrypoint (str): The uri relative path.
            methods (tuple): A tuple of valid method strings.
            callable_obj (callable): The callable object.

        Returns:
            The Router instance (for chaining purposes).

        Raises:
            RouteError, for missing routing params or invalid callable
            object type.
        """

        router_obj = Route.wrap_callable(
            uri=entrypoint,
            methods=methods,
            callable_obj=callable_obj
        )

        if router_obj.is_valid:
            self._routes.add(router_obj)
            return self
        
        raise RouteError(  # pragma: no cover
            "Missing params: methods: {} - entrypoint: {}".format(
                methods, entrypoint
            )
        )

    def mount(self, app=None):
        """Mounts all registered routes to a bottle.py application instance.

        Args:
            app (instance): A `bottle.Bottle()` application instance.

        Returns:
            The Router instance (for chaining purposes).
        """
        for endpoint in self._routes:
            endpoint.register_app(app)

        return self

    def __repr__(self):  # pragma: no cover
        return 'Router object: total {} routes'.format(len(self))

    def __len__(self):
        return len(self._routes)

    def __iter__(self):  # pragma: no cover
        for route in self._routes:
            yield route



# -*- coding: utf-8 -*-
"""Unit Tests for `bottle_neck.handlers` module.
"""


import pytest
import bottle_neck.response as response


def test_ws_response_init_pass():
    """Test `bottle_neck.response.WSResponse` init pass.
    """
    assert response.WSResponse(200, 1)


def test_ws_response_init_fail():
    """Test `bottle_neck.response.WSResponse` init pass.
    """
    with pytest.raises(response.WSResponseError):
        assert response.WSResponse(209, 1)


def test_ws_response_ok():
    """Test `bottle_neck.response.WSResponse.ok` method.
    """
    assert response.WSResponse.ok(data={"count": 1})


def test_ws_response_bad_request():
    """Test `bottle_neck.response.WSResponse.bad_request` method.
    """
    assert response.WSResponse.bad_request(errors=[1, 2])


def test_ws_response_created():
    """Test `bottle_neck.response.WSResponse.created` method.
    """
    assert response.WSResponse.created(data={})


def test_ws_response_not_modified():
    """Test `bottle_neck.response.WSResponse.not_modified` method.
    """
    assert response.WSResponse.not_modified()



# -*- coding: utf-8 -*-
"""Unit Tests for `bottle_neck.handlers` module.
"""

__author__ = 'Papavassiliou Vassilis'


import pytest
import bottle_neck.cbv as handler


@pytest.fixture(scope="module")
def mock_handler():
    """Pytest fixture for `bottle_neck.handlers.BaseHandler` subclass.
    """

    class MockHandler(handler.BaseHandler):
        """Web handler fixture class.
        """
        cors_enabled = True
        base_endpoint = '/mock'

        def get(self, mock_id):
            return mock_id

        @handler.route_method('GET', extra_part=True)
        @handler.plugin_method('mock_plugin')
        def version(self):
            return 0, 0, 1

    return MockHandler


@pytest.fixture(scope="module")
def mock_plugin():
    """Pytest fixture for `bottle_neck.handlers.BasePlugin` subclass.
    """

    class MockPlugin(handler.BaseHandlerPlugin):

        def apply(self, *args, **kwargs):
            try:
                kwargs['pk'] = int(kwargs['pk'])
            except ValueError:
                kwargs['pk'] = -1

            return self._wrapped(*args, **kwargs)

    return MockPlugin


def test_handler_route_method_function_pass():
    """Test `bottle_neck.handlers.route_method` function pass.
    """
    class TestCls(object):
        @handler.route_method('POST', extra_part=True)
        def fn(self):
            pass

    assert TestCls.fn.http_method == 'POST' \
        and TestCls.fn.url_extra_part is 'fn'


def test_handler_route_method_function_fail():
    """Test `bottle_neck.handlers.route_method` function fail.
    """

    with pytest.raises(handler.HandlerHTTPMethodError):
        class TestCls(object):
            @handler.route_method('PORT', extra_part=True)
            def fn(self):
                pass

        assert TestCls


def test_handler_plugin_method_function_pass():
    """Test `bottle_neck.handlers.route_method` function pass.
    """
    class TestCls(object):
        @handler.plugin_method('log', 'auth')
        def fn(self):
            pass

    assert TestCls.fn.log is TestCls.fn.auth is True


def test_handler_singleton(mock_handler):
    """Test `bottle_neck.handlers.BaseHandler` subclass for singleton pattern.
    """
    assert mock_handler() == mock_handler()


def test_handler_cls_attrs(mock_handler):
    """Test `bottle_neck.handlers.BaseHandler` subclass for init class attrs.
    """
    ch_attrs = ('plugins', 'global_plugins', 'base_endpoint', 'cors_enabled')

    assert set(ch_attrs).issubset(set(dir(mock_handler)))


def test_handler_cls_add_plugin_function_based(mock_handler):
    """Test `bottle_neck.handlers.BaseHandler` explicit adding a plugin.
    """
    def kwargs_plugin(handler):
        """Testing function-based plugin.
        """
        def _decorator(*args, **kwargs):  # pragma: no cover
            return handler(*args, **kwargs)

        return _decorator

    mock_handler.add_plugin(
        global_scope=False,
        plugin_callables=[kwargs_plugin]
    )

    assert 'kwargs_plugin' in mock_handler.plugins


def test_handler_add_plugin_class_based(mock_handler, mock_plugin):
    """Test `bottle_neck.handlers.BaseHandler.add_plugin` method.
    """
    mock_handler.add_plugin(
        global_scope=True,
        plugin_callables=[mock_plugin]
    )

    assert 'mock_plugin' in mock_handler.global_plugins
    assert mock_plugin in mock_handler.global_plugins.values()


def test_handler_register_class_fail(mock_app):
    """Test `bottle_neck.handlers.BaseHandler.register_app` method error
    handling.
    """

    with pytest.raises(handler.HandlerError):
        handler.BaseHandler.register_app(mock_app)


def test_handler_register_class_pass(mock_app, mock_handler):
    """Test `bottle_neck.handlers.BaseHandler.register_app` method.
    """

    mock_handler.register_app(mock_app)

    assert all(['/mock' in i.rule for i in mock_app.routes])



# -*- coding: utf-8 -*-
"""Bottle.py API development utilities.

Provides some extra functionality for developing web API's with `bottle.py`.
"""

from __future__ import absolute_import

__author__ = 'pav'
__date__ = '2015-2-3'
__all__ = ['cors_enable_hook', 'strip_path_hook', 'paginator']

from bottle_neck import __version__
from collections import OrderedDict
import bottle
import math

version = tuple(map(int, __version__.split('.')))


def cors_enable_hook():
    bottle.response.headers['Access-Control-Allow-Origin'] = '*'
    bottle.response.headers['Access-Control-Allow-Headers'] = \
        'Authorization, Credentials, X-Requested-With, Content-Type'
    bottle.response.headers['Access-Control-Allow-Methods'] = \
        'GET, PUT, POST, OPTIONS, DELETE'


def strip_path_hook():
    """Ignore trailing slashes.
    """
    bottle.request.environ['PATH_INFO'] = \
        bottle.request.environ['PATH_INFO'].rstrip('/')


def paginator(limit, offset, record_count, base_uri, page_nav_tpl='&limit={}&offset={}'):
    """Compute pagination info for collection filtering.

    Args:
        limit (int): Collection filter limit.
        offset (int): Collection filter offset.
        record_count (int): Collection filter total record count.
        base_uri (str): Collection filter base uri (without limit, offset)
        page_nav_tpl (str): Pagination template.

    Returns:
        A mapping of pagination info.
    """

    total_pages = int(math.ceil(record_count / limit))

    next_cond = limit + offset <= record_count
    prev_cond = offset >= limit

    next_page = base_uri + page_nav_tpl.format(limit, offset + limit) if next_cond else None

    prev_page = base_uri + page_nav_tpl.format(limit, offset - limit) if prev_cond else None

    return OrderedDict([
        ('total_count', record_count),
        ('total_pages', total_pages),
        ('next_page', next_page),
        ('prev_page', prev_page)
    ])
